import json
from collections import OrderedDict
from functools import partial
from typing import Any, Dict, List, MutableMapping, Tuple

import click
import pendulum
import yaml
from yaml import SafeDumper, ScalarNode, SequenceNode
from yaml.representer import SafeRepresenter

from esque.resources.topic import Topic, TopicDiff, Watermark

MILLISECONDS_PER_YEAR = 1000 * 3600 * 24 * 365


def _indent(level: int):
    return "  " * level


def pretty(value, *, break_lists=False) -> str:
    if isinstance(value, dict):
        value_str = pretty_dict(value)
    elif isinstance(value, (list, tuple)):
        value_str = pretty_list(value, break_lists=break_lists)
    elif type(value) in TYPE_MAPPING:
        value_str = TYPE_MAPPING[type(value)](value)
    else:
        value_str = str(value)

    if len(value_str.strip()) == 0:
        return "''"
    return value_str


def pretty_list(
    ugly_list: List[Any], *, break_lists=False, list_separator: str = ", ", broken_list_separator: str = "- "
) -> str:
    if len(ugly_list) == 0:
        return "[]"

    list_output = [pretty(value) for value in ugly_list]

    if any("\n" in list_element for list_element in list_output):
        break_lists = True

    if break_lists:
        sub_elements = ("\n  ".join(elem.splitlines(keepends=False)) for elem in list_output)
        return broken_list_separator + f"\n{broken_list_separator}".join(sub_elements)
    else:
        return list_separator.join(list_output)


def pretty_dict(d: MutableMapping[str, Any]) -> str:
    if len(d) == 0:
        return "{}"

    max_col_length = max([len(str(key)) for key in d.keys()])

    lines = []
    for key, value in d.items():
        if is_scalar(value):
            row = f"{pretty(key)}:".ljust(max_col_length + 2)

            unit = str(key).split(".")[-1]
            row += get_value(unit, value)
        else:
            row = f"{pretty(key)}:\n  "
            sub_lines = pretty(value).splitlines(keepends=False)
            row += "\n  ".join(sub_lines)

        if key in STYLE_MAPPING:
            row = STYLE_MAPPING[key](row)

        lines.append(row)

    return "\n".join(lines)


def get_value(unit: str, value: Any) -> str:
    if isinstance(value, str) and " -> " in value:
        values = value.split(" -> ")
        return (
            click.style(get_value(unit, values[0]), fg="red")
            + " -> "
            + click.style(get_value(unit, values[1]), fg="green")
        )

    if unit in CONVERSION_MAPPING:
        return f"{CONVERSION_MAPPING[unit](value)}"
    return pretty(value)


def is_scalar(value: Any) -> bool:
    return (value is None) or (not isinstance(value, (list, dict, OrderedDict)))


def pretty_pendulum(value: pendulum.DateTime) -> str:
    return value.to_datetime_string()


def pretty_float(value: float) -> str:
    return f"{value:0.2f}"


def pretty_int(value: int) -> str:
    return f"{value}"


def pretty_bytes(value: bytes) -> str:
    return value.decode("UTF-8")


def pretty_duration(value: Any, *, multiplier: int = 1) -> str:
    if not value:
        return ""

    if type(value) != int:
        value = int(value)

    value *= multiplier

    # Fix for conversion errors of ms > C_MAX_INT in some internal lib
    if value > MILLISECONDS_PER_YEAR:
        value = int(value / MILLISECONDS_PER_YEAR)
        return pendulum.duration(years=value).in_words()

    return pendulum.duration(milliseconds=value).in_words()


def pretty_topic_diffs(topics_config_diff: Dict[str, TopicDiff]) -> str:
    output = []
    for name, diff in topics_config_diff.items():
        config_diff_attributes = {}
        for attr, old, new in diff.changes():
            config_diff_attributes[attr] = f"{old} -> {new}"
        output.append({click.style(name, bold=True, fg="yellow"): {"Config Diff": config_diff_attributes}})

    return pretty({"Configuration changes": output})


def pretty_new_topic_configs(new_topics: List[Topic]) -> str:
    new_topic_configs = []
    for topic in new_topics:
        new_topic_config = {
            "num_partitions: ": topic.num_partitions,
            "replication_factor: ": topic.replication_factor,
            "config": topic.config,
        }
        new_topic_configs.append({click.style(topic.name, bold=True, fg="green"): new_topic_config})

    return pretty({"New topics to create": new_topic_configs})


def pretty_unchanged_topic_configs(new_topics: List[Topic]) -> str:
    new_topic_configs = []
    for topic in new_topics:
        new_topic_config = {
            "num_partitions: ": topic.num_partitions,
            "replication_factor: ": topic.replication_factor,
            "config": topic.config,
        }
        new_topic_configs.append({click.style(topic.name, bold=True, fg="blue"): new_topic_config})

    return pretty({"No changes": new_topic_configs})


def pretty_size(value: Any) -> str:
    if type(value) != int:
        value = int(value)
    units = [
        ("Eib", 1024 ** 6),
        ("Pib", 1024 ** 5),
        ("Tib", 1024 ** 4),
        ("Gib", 1024 ** 3),
        ("Mib", 1024 ** 2),
        ("kib", 1024 ** 1),
        ("b", 1024 ** 0),
    ]
    for sign, size in units:
        if value >= size:
            return f"{pretty_float(value / size)} {sign}"


def bold(s: str) -> str:
    return click.style(s, bold=True)


def blue_bold(s: str) -> str:
    return bold(click.style(s, fg="blue"))


def green_bold(s: str) -> str:
    return bold(click.style(s, fg="green"))


def red_bold(s: str) -> str:
    return bold(click.style(s, fg="red"))


STYLE_MAPPING = {
    "topic": green_bold,
    "cleanup.policy": bold,
    "flush.ms": bold,
    "delete.retention.ms": bold,
    "retention.ms": bold,
    "low_watermark": blue_bold,
    "high_watermark": blue_bold,
    "member_id": bold,
}

CONVERSION_MAPPING = {
    "ms": pretty_duration,
    "seconds": partial(pretty_duration, multiplier=1000),
    "bytes": pretty_size,
}

TYPE_MAPPING = {
    pendulum.DateTime: pretty_pendulum,
    float: pretty_float,
    int: pretty_int,
    dict: pretty_dict,
    OrderedDict: pretty_dict,
    list: pretty_list,
    bytes: pretty_bytes,
}


def tuple_representer(dumper: SafeRepresenter, data: Watermark) -> SequenceNode:
    return dumper.represent_list(list(data))


def watermark_representer(dumper: SafeRepresenter, data: Watermark) -> SequenceNode:
    return dumper.represent_list(list(data))


def bytes_representer(dumper: SafeRepresenter, data: bytes) -> ScalarNode:
    return dumper.represent_str(data.decode("UTF-8"))


SafeDumper.add_representer(Tuple, tuple_representer)
SafeDumper.add_representer(bytes, bytes_representer)
SafeDumper.add_representer(Watermark, watermark_representer)


class BytesEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, bytes):
            return obj.decode("UTF-8")
            # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def format_output(output: Any, output_format: str) -> str:
    if output_format == "yaml":
        return yaml.dump(output, default_flow_style=False, sort_keys=False, Dumper=yaml.SafeDumper)
    elif output_format == "json":
        return json.dumps(output, indent=4, cls=BytesEncoder)
    elif isinstance(output, list):
        return pretty_list(output, break_lists=True, broken_list_separator="")
    else:
        return pretty(output, break_lists=True)



import datetime


def zero_or_more(element):
    return "(" + element + ")*"


def one_or_more(element):
    return "(" + element + ")+"


def zero_or_one(element):
    return "(" + element + ")?"


def exactly(element, count):
    return "(" + element + "){" + str(count) + "}"


def either_of(iterable):
    return "((" + ")|(".join(iterable) + "))"


def is_float(input_string: str):
    try:
        float(input_string)
        return True
    except ValueError:
        return False


def is_int(input_string: str):
    try:
        int(input_string)
        return True
    except ValueError:
        return False


def is_any_numeric_type(input_string: str):
    return is_float(input_string) or is_int(input_string)


def is_date_string(input_string: str):
    try:
        datetime.datetime.strptime(input_string, "%Y-%m-%d")
        return True
    except ValueError:
        return False


def is_date_time_string(input_string: str):
    try:
        datetime.datetime.strptime(input_string, "%Y-%m-%dT%H:%M:%S")
        return True
    except ValueError:
        return False


def is_boolean(input_string: str):
    return to_boolean(input_string) is not None


def to_boolean(input_string: str):
    """Returns True, False or None, based on the contents of the input_string (any form of capitalization is allowed)"""
    if input_string.lower() == "true":
        return True
    elif input_string.lower() == "false":
        return False
    else:
        return None


def string_like(haystack: str, needle: str):
    if needle.startswith("%") and needle.endswith("%"):
        return haystack.find(needle.replace("%", "")) != -1
    elif needle.startswith("%"):
        return haystack.endswith(needle.replace("%", ""))
    elif needle.endswith("%"):
        return haystack.startswith(needle.replace("%", ""))
    else:
        return haystack == needle.replace("%", "")


def string_not_like(haystack: str, needle: str):
    return not string_like(haystack, needle)



import json
import pathlib
from abc import abstractmethod
from glob import glob
from typing import Any, Iterable, List, NamedTuple, Optional

import click
from avro.schema import RecordSchema
from confluent_kafka.cimpl import Message


class MessageHeader(NamedTuple):
    key: str
    value: Optional[str]


class DecodedMessage(NamedTuple):
    key: str
    value: Optional[str]
    partition: int
    offset: int
    timestamp: str
    headers: List[MessageHeader] = []


class KafkaMessage(NamedTuple):
    key: Any
    value: Any
    partition: int
    key_schema: Optional[RecordSchema] = None
    value_schema: Optional[RecordSchema] = None
    headers: List[MessageHeader] = []


class GenericWriter:
    @abstractmethod
    def write_message(self, message: Message):
        pass


class FileHandler:
    def __init__(self, directory: pathlib.Path, partition: Optional[str] = None):
        self.directory = directory
        self.partition = partition
        self.open_mode = "w+"
        self.file = None

    def __enter__(self):
        self.init_destination_directory()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.close()

    def init_destination_directory(self):
        partition = "any" if self.partition is None else self.partition
        partition_file = self.directory / f"partition_{partition}"
        self.file = partition_file.open(self.open_mode)


class StdOutWriter(GenericWriter):
    def write_message(self, message: Message):
        click.echo(serialize_message(message))


class FileWriter(GenericWriter, FileHandler):
    def __init__(self, directory: pathlib.Path, partition: Optional[str] = None):
        super().__init__(directory, partition)
        self.open_mode = "w+"

    def write_message(self, message: Message):
        pass


class FileReader(FileHandler):
    def __init__(self, directory: pathlib.Path, partition: Optional[str] = None):
        super().__init__(directory, partition)
        self.open_mode = "r"

    def read_message_from_file(self) -> Iterable[KafkaMessage]:
        pass


class PlainTextFileWriter(FileWriter):
    def write_message(self, message: Message):
        self.file.write(serialize_message(message) + "\n")


class PlainTextFileReader(FileReader):
    def read_message_from_file(self) -> Iterable[KafkaMessage]:
        for line in self.file:
            yield deserialize_message(line)


def decode_message(message: Message) -> DecodedMessage:
    decoded_key = None if message.key() is None else message.key().decode("utf-8")
    decoded_value = None if message.value() is None else message.value().decode("utf-8")
    headers = []
    if message.headers():
        for header_key, header_value in message.headers():
            headers.append(MessageHeader(key=header_key, value=header_value.decode("utf-8") if header_value else None))
    return DecodedMessage(
        key=decoded_key,
        value=decoded_value,
        partition=message.partition(),
        offset=message.offset(),
        timestamp=str(message.timestamp()),
        headers=headers,
    )


def serialize_message(message: Message):
    decoded_message = decode_message(message)
    serializable_message = {
        "key": decoded_message.key,
        "value": decoded_message.value,
        "partition": decoded_message.partition,
        "offset": decoded_message.offset,
        "headers": [
            {"key": header_element.key, "value": header_element.value} for header_element in decoded_message.headers
        ],
    }
    return json.dumps(serializable_message)


def deserialize_message(message_line: str) -> KafkaMessage:
    json_record = json.loads(message_line)
    key = None if "key" not in json_record or json_record["key"] is None else json_record["key"]
    value = None if "value" not in json_record or json_record["value"] is None else json_record["value"]
    partition = json_record.get("partition", 0)
    headers = []
    if json_record["headers"]:
        for header_item in json_record["headers"]:
            header_key = header_item["key"]
            header_value = header_item["value"] if "value" in header_item else None
            if header_key:
                headers.append(MessageHeader(header_key, header_value if header_value else None))
    return KafkaMessage(key=key, value=value, partition=partition, headers=headers)


def get_partitions_in_path(input_directory: pathlib.Path) -> Iterable[str]:
    path_list = glob(str(input_directory / "partition_*"))
    partitions = []
    for partition_path in path_list:
        partitions.append(partition_path.rsplit("_", 1)[1])
    return partitions



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['esque',
 'esque.cli',
 'esque.clients',
 'esque.config',
 'esque.controller',
 'esque.messages',
 'esque.resources',
 'esque.ruleparser',
 'esque.validation']

package_data = \
{'': ['*'], 'esque.validation': ['schemas/*']}

install_requires = \
['avro-python3==1.8.2',
 'click>=7.0,<8.0',
 'confluent-kafka[avro]>=1.2,<2.0',
 'fastavro>=0.22.5,<0.23.0',
 'pendulum>=2.0,<3.0',
 'pykafka>=2.8,<3.0',
 'pyyaml>=5.1,<6.0',
 'requests>=2.22,<3.0',
 'toml>=0.10.0,<0.11.0',
 'yamale>=2.0,<3.0']

entry_points = \
{'console_scripts': ['esque = esque.cli.commands:esque']}

setup_kwargs = {
    'name': 'esque',
    'version': '0.2.0b3',
    'description': 'esque - an operational kafka tool.',
    'long_description': '# esque - an operational Kafka tool\n\n[![pypi Version](https://img.shields.io/pypi/v/esque.svg)](https://pypi.org/project/esque/)\n[![Python Versions](https://img.shields.io/pypi/pyversions/esque.svg)](https://pypi.org/project/esque/)\n![Build Status](https://github.com/real-digital/esque/workflows/Style,%20Unit%20And%20Integration%20Tests/badge.svg)\n[![Coverage Status](https://coveralls.io/repos/github/real-digital/esque/badge.svg?branch=master)](https://coveralls.io/github/real-digital/esque?branch=master)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\nIn the Kafka world nothing is easy, but `esque` (pronounced *esk*) is an attempt at it.\n\n`esque` is a user-centric command line interface for Kafka administration.\n\n## Why should you care?\n\nSome stuff is hard, and that is okay, but listing your kafka topics shouldn\'t be.\n\nWhile adopting kafka at real.digital we noticed the immense entry barrier it poses to newcomers.\nWe can\'t recount how often we wrote Slack messages asking for the script to check the\nstatus of topics or consumer groups. This is partly (but not only) due to a\nfragmented and unclear definition of tooling and APIs for kafka.\nIn a wide array of administration tools, `esque` distances itself by striving to provide Kafka Ops for Humans, in a usable and natural way.\n\nWe feel that the goal of `esque` embodies the principle: “**keep easy things easy, and make hard things possible**”.\n\n## Principles\n\n* batteries included\n* feature rich\n* robust\n* insightful\n* by engineers for engineers\n\n## Feature Overview\n\n* Support for any type of Kafka deployment >1.2\n* Display Resources (Topics, Consumer Groups, Brokers)\n* Get detailed Overviews of Resources (Topics, Consumer Groups, Brokers)\n* Create/Delete Topics\n* Edit Topic Configurations\n* Edit Consumer Offset for Topics\n* SASL/SSL Support out of the box\n* Consume and Produce to and from Avro and Plaintext Topics (including Avro Schema Resolution from Schema Registry)\n* Context Switch (Easily Switch between pre-defined Clusters)\n* Kafka Ping (Test roundtrip time to your kafka cluster)\n\n## Command Overview\n\n```bash\n$ esque\n\nUsage: esque [OPTIONS] COMMAND [ARGS]...\n\n  esque - an operational kafka tool.\n\n  In the Kafka world nothing is easy, but esque (pronounced esk) is an\n  attempt at it.\n\nOptions:\n  --recreate-config  Overwrites the config with the sample config.\n  --version          Show the version and exit.\n  -v, --verbose      Return stack trace on error.\n  --no-verify        Skip all verification dialogs and answer them with yes.\n  --help             Show this message and exit.\n\nCommands:\n  apply     Apply a set of topic configurations.\n  config    Configuration-related options.\n  consume   Consume messages from a topic.\n  create    Create a new instance of a resource.\n  ctx       List contexts and switch between them.\n  delete    Delete a resource.\n  describe  Get detailed information about a resource.\n  edit      Edit a resource.\n  get       Get a quick overview of different resources.\n  ping      Test the connection to the kafka cluster.\n  produce   Produce messages to a topic.\n  set       Set resource attributes.\n\n```\n\n## Installation and Usage\n\n### Installation\n\n`esque` is available at [pypi.org](https://pypi.org/project/esque/) and can be installed with `pip install esque`. `esque` requires Python 3.6+ to run.\n\n#### SASL Support\n\nWhen your cluster is secured with SASL authentication, you\'ll need to install our fork of pykafka since pykafka itself\ndoesn\'t support it. We\'ve opened a pull request https://github.com/Parsely/pykafka/pull/972 but at the time of\nwriting it hasn\'t been merged yet.\n\n```bash\npip install -U git+https://github.com/real-digital/pykafka.git@feature/sasl-scram-support\n```\n\n`esque` will also prompt you with the above command as soon as you need it in case you\'re not sure if you actually do.\n\n### Autocompletion\n\nThe autocompletion scripts for `bash` and `zsh` can be generated by running `esque config autocomplete`.\n\n### Usage\n\n#### Config Definition\n\nWhen starting `esque` for the first time the following message will appear:\n\n```bash\nNo config provided in ~/.esque\nShould a sample file be created in ~/.esque [y/N]:\n```\n\nWhen answering with `y` `esque` will copy over the [sample config](https://github.com/real-digital/esque/blob/master/esque/config/sample_config.yaml) to `~/.esque/esque_config.yaml`.\nAfterwards you can modify that file to fit your cluster definitions.\n\nAlternatively might just provide a config file following the sample config\'s file in that path.\n\n##### Config Example\n\n```yaml\nversion: 1\ncurrent_context: local\ncontexts:\n  # This context corresponds to a local development cluster\n  # created by docker-compose when running esque from the host machine.\n  local:\n    bootstrap_servers:\n      - localhost:9092\n    security_protocol: PLAINTEXT\n    schema_registry: http://localhost:8081\n    default_values:\n      num_partitions: 1\n      replication_factor: 1\n```\n\n#### Config file for "apply" command\n\nThe config for the apply command has to be a yaml file and\nis given with the option -f or --file.\n\nIn the current version only topic configurations can be\nchanged and specified.\n\nIt has to use the same schema, which is used\nfor the following example:\n\n```yaml\ntopics:\n  - name: topic_one\n    replication_factor: 3\n    num_partitions: 50\n    config:\n      cleanup.policy: compact\n  - name: topic_two\n    replication_factor: 3\n    num_partitions: 50\n    config:\n      cleanup.policy: compact\n```\n\n## Development\n\nTo setup your development environment, make sure you have at least Python 3.6 & [poetry](https://github.com/sdispater/poetry) installed, then run\n\n```bash\npoetry install\npoetry run pip install -U git+https://github.com/real-digital/pykafka.git@feature/sasl-scram-support\npoetry shell\n```\n\n### Pre Commit Hooks\n\nTo install pre commit hooks run:\n\n```bash\npip install pre-commit\npre-commit install\npre-commit install-hooks\n```\n\n### Run tests\n\n#### Integration Tests\n\nesque comes with a docker-compose based kafka stack which you can start up with `make test-suite`.\n\nYou can then run the integration tests against this stack with `pytest tests/ --integration --local`.\n\nAlternatively you can go the fast way and just run the whole stack + integration tests in docker:\n\n```bash\nmake integration-test\n```\n\n#### Unit Tests\n\nIf you only want the unit tests, just run:\n\n```bash\nmake test\n```\n\n## Alternatives\n\n- [LinkedIn KafkaTools](https://github.com/linkedin/kafka-tools)\n- [PyKafka Tools](https://github.com/Parsely/pykafka/blob/master/pykafka/cli/kafka_tools.py)\n- [Official Kafka Scripts](https://github.com/apache/kafka/tree/trunk/bin)\n- [kafkacat](https://github.com/edenhill/kafkacat)\n- [kaf](https://github.com/birdayz/kaf)\n',
    'author': 'real.digital',
    'author_email': 'opensource@real-digital.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/real-digital/esque',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
